#define MODULE    "BANDWIDTH"
#define MODULE_ID "@(#)$Id: bandwidth.c,v 1.11 2001/06/18 08:39:51 hob Exp $"

/* Copyright Notice ----------------------------------------------------------
//
// bandwidth.c: MPI point-to-point performance benchmark
//
// Copyright (C) 2001  Scali AS   
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// ---------------------------------------------------------------------------
*/

/* Module --------------------------------------------------------------------
// 
// $RCSfile: bandwidth.c,v $
// 
// CREATED
//   Author: hob (Hakon Ording Bugge)
//   Date:   2001/02/28 16:28:41
// 
// LAST CHANGED
//   $Author: hob $
//   $Date: 2001/06/18 08:39:51 $
// 
// DESCRIPTION
//   This program serves the purpose of assessing point-to-point performance
/    of an MPI implementation.
//   
// WARNING
//   Scali considers this program not to be the *best* program for measuring
//   performance, but it is included for its simplicity. Scali considers
//   collective operations to be more valuable in assesing a clusters'
//   performance, hence, use Pallas PMB or mpptest instead.
//   
//   
// -------------------------------------------------------------------------*/

/* Dependencies ------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "mpi.h"

/* Constants ---------------------------------------------------------------*/

#define TAG 123

#define DO_PING_PING 1
#define DO_PING_PONG 2
#define DO_EXCHANGE  4

/* Typedefs ----------------------------------------------------------------*/

typedef void (*bench_rout_t)(int, char *, int);
typedef struct {
      bench_rout_t bench;
      char *mnem;
      int scale;
      int mask;
} bench_desc_t;
      
/* Globals -----------------------------------------------------------------*/

int me, nproc;
int touch_tx_initially=1;
int touch_tx_always=0;
int touch_rx=0;
const int min_iter=2;
double tgoal=0.1, tres;

void ping_ping(int iter, char *buf, int len) {
   MPI_Status sts;
   int i;

   for (i=0; i < iter; i++) {
      if (me == 0) {
         if (touch_tx_always) memset((void *)buf, 0, len);
	 MPI_Send(buf, len, MPI_BYTE, 1, TAG, MPI_COMM_WORLD);
      } else {
	 MPI_Recv(buf, len, MPI_BYTE, 0, TAG, MPI_COMM_WORLD, &sts);
	 if (touch_rx) memcmp((void *)buf, (void *)buf, len);
      }
   }
}

void ping_pong(int iter, char *buf, int len) {
   MPI_Status sts;
   int i;

   for (i=0; i < iter; i++) {
      if (me == 0) {
         if (touch_tx_always) memset((void *)buf, 0, len);
	 MPI_Send(buf, len, MPI_BYTE, 1, TAG, MPI_COMM_WORLD);
	 MPI_Recv(buf, len, MPI_BYTE, 1, TAG, MPI_COMM_WORLD, &sts);
	 if (touch_rx) memcmp((void *)buf, (void *)buf, len);
      } else {
	 MPI_Recv(buf, len, MPI_BYTE, 0, TAG, MPI_COMM_WORLD, &sts);
	 if (touch_rx) memcmp((void *)buf, (void *)buf, len);
         if (touch_tx_always) memset((void *)buf, 0, len);
	 MPI_Send(buf, len, MPI_BYTE, 0, TAG, MPI_COMM_WORLD);
      }
   }
}

void exchange(int iter, char *buf, int len) {
   MPI_Status sts;
   int i;

   for (i=0; i < iter; i++) {
      if (touch_tx_always) memset((void *)buf, 0, len);
      MPI_Sendrecv(buf, len, MPI_BYTE, me^1, TAG,
		   buf, len, MPI_BYTE, me^1, TAG,
		   MPI_COMM_WORLD, &sts);
      if (touch_rx) memcmp((void *)buf, (void *)buf, len);
   }
}

void print_header(char *str) {
   int i;
   
   if (!me) {
      printf("Benchmark %s\n", str);
      for (i=0; i < strlen(str)+strlen("Benchmark "); ++i) printf("=");
      printf("\n");

      printf("%14s %14s %14s %14s %14s\n",
	     "lenght", "iterations", "elapsed time", "transfer rate", "latency");
      printf("%14s %14s %14s %14s %14s\n",
	     "(bytes)", "(count)", "(seconds)", "(Mbytes/s)", "(usec)");
      printf("--------------------------------------------------------------------------\n");
   }
}

void measure(bench_rout_t rout, int scale, int sz) {
   int i, iter;
   char *p = (char *)malloc(sz);
   double t, tmin, s;
   
   if (touch_tx_initially) memset((void *)p, 0, sz);
   
   /* First, quick evaluation to find the correct no of iterations */
   for (iter=1; 1; iter+= iter) {
      tmin = 9999.;
      for (i=0; i < 3; ++i) { /* minimum of 3 times seems good */
	 rout(1, p, sz);
	 t = MPI_Wtime();
	 rout(iter, p, sz);
	 t = MPI_Wtime() - t;
	 MPI_Allreduce(&t, &s, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
	 t = s / (double)nproc;
      }
      tmin =  (t < tmin) ? t : tmin;
      if (tmin > 10.0*tres) break;
   }

   /* compute estimated no iterations to reach tgoal */
   iter = iter * (tgoal / tmin);
   if (iter < min_iter) iter = min_iter; /* at least min_iter iterations */
   
   /* measure with one dry run first to get things settled in cache */
   for (i=0; i < 2; ++i) {
      t = MPI_Wtime();
      rout((i == 0) ? 1 : iter, p, sz);
      t = MPI_Wtime() - t;
   }

   MPI_Allreduce(&t, &s, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
   t = s / (double)nproc;

   if (t <= 0.0) t = tres;  /* avoid crash ... */
   
   if (me == 0) {
      printf("%14d %14d %14.3f %14.1f %14.1f\n",
	     sz, iter, t,
	     sz*(double)iter*(double)scale*1e-6/t,
	     t*1e6/(double)iter/(double)scale);
   }

   free(p);
}

      
bench_desc_t bench[] = { 
   { ping_ping, "ping-ping",              1, DO_PING_PING },
   { ping_pong, "ping-pong",              2, DO_PING_PONG },
   { exchange,  "exchange(MPI_Sendrecv)", 2, DO_EXCHANGE  },
   { NULL,      "none",      0 },
};


void do_usage(void) {
   if (!me) {
      printf("Usage: mpimon <mpimon switches> bandwidth [-b i|o|x] [-h] [-m min] [-M max] [-t tgoal] [-V] [-z mask] -- hosta hostb\n");
   }
}

void do_help(void) {
  if (!me) {
    do_usage();
    printf("\
      -b <i|o|x>   : Benchmark to run, pIng-pIng, ping-pOng, or eXchange.\n\
                     Default all.\n\
      -h           : Prints this message\n\
      -m <int>     : Minimum message length. Default 0.\n\
      -M <int>     : Maximum message length. Default 16M.\n\
      -t <float>   : Specifies the goal for the elapsed time used per test.\n\
                     Default %.1f\n\
      -V           : Prints version.\n\
      -z <mask>    : Buffer initialization/touch. Default 1.\n\
          mask&1   : Initial initialization of send buffer.\n\
          mask&2   : Initialization of send buffer before every send.\n\
          mask&4   : Read receive buffer after message has been received.\n\
      ", tgoal);
  }
}

      
void do_exit(int exit_code) {
   
   if (exit_code) {
      MPI_Abort(MPI_COMM_WORLD, exit_code);
   }
   MPI_Finalize();
   exit(exit_code);
}
      
   
int main(int argc, char **argv) {
   extern char *optarg;
   int c, j, inc, bench_mask=0;
   size_t sz, msg_min=0, msg_max=16*1024*1024;
   double t1, t2, best_tres;

   MPI_Init(&argc, &argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&me);
   MPI_Comm_size(MPI_COMM_WORLD, &nproc);

   if (nproc != 2) {
      if (!me) fprintf(stderr, "Expected two processes, found %d\n",
		       nproc);
      do_exit(1);
   }
   
   while ((c = getopt(argc, argv, "b:hm:M:t:Vz:")) != EOF) {
      switch (c) {
	 case 'b':
	    switch (*optarg) {
	       case 'i':
		  bench_mask|= DO_PING_PING;
		  break;
		  
	       case 'o':
		  bench_mask|= DO_PING_PONG;
		  break;
		  
	       case 'x':
		  bench_mask|= DO_EXCHANGE;
		  break;
		  
	       case '?':
		  do_usage();
		  do_exit(3);
		  break;
	    }
	    break;
	       
	 case 'h':
	    do_help();
	    do_exit(0);
	    break;

	 case 'm':
	    msg_min = atoi(optarg);
	    break;
	    
	 case 'M':
	    msg_max = atoi(optarg);
	    break;
	    
	 case 't':
	    tgoal = atof(optarg);
	    break;

	 case 'V':
	    if (me == 0) printf("%s\n", MODULE_ID);
	    break;

	 case 'z':
	    touch_tx_initially = atoi(optarg) & 1;
	    touch_tx_always    = atoi(optarg) & 2;
	    touch_rx           = atoi(optarg) & 4;
	    break;

	 case '?':
	    do_usage();
	    do_exit(2);
      }
   }
   
   /* if no benchmarks have been specified, we default to all */
   if (!bench_mask) bench_mask = DO_PING_PING|DO_PING_PONG|DO_EXCHANGE;
   
   /* find clock resolution */
   for (j=0, best_tres=1e6; j<10; ++j) {
      for (t1=MPI_Wtime(); (t2=MPI_Wtime()) == t1;);
      t2 = t2 - t1;
      MPI_Allreduce(&t2, &tres, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
      tres = tres / (double)nproc;
      if (tres < best_tres) best_tres=tres;
   }

   tres = best_tres;
   
   if (me == 0) {
      printf("Resolution (usec): %f\n", tres*1e6);
   }
   
   for (j=0; bench[j].bench; ++j) {

      if (!(bench_mask & bench[j].mask)) continue;
	  
      print_header(bench[j].mnem);
      inc = msg_min >> 2;
      for (sz=msg_min; sz <= msg_max; sz+=inc) {
	 if (sz < 4) {
	    inc = 1;
	 } else {
	    if (!(sz & inc)) inc+=inc; /* Got this? */
	 }

	 measure(bench[j].bench, bench[j].scale, sz);
      }
      if (!me) printf("\n\n");
   }
   do_exit(0);
   
   return 0;
}

	 
	 
	 
	 
	 
	 

