/* pingpong - measure effective bandwidth and latency
   this is a variant to test for non-contignous data 
   (using MPI vector type) */

#include <stdio.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <errno.h>

#include "mpi.h"

int main(int argc, char **argv) {
    MPI_Status status;
    MPI_Datatype noncontig,contig;

    int myrank, mysize;
    int size, bls, ble, count, stride, blocklen, repeats, i,j, total_size, f;
    int *buffer, *recv;
    double bandwidth, latency, wtime, start;
    char c;
    
    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
    MPI_Comm_size(MPI_COMM_WORLD, &mysize);

#define DEF_BL_START 8
#define DEF_BL_END 131072
#define DEF_SIZE 262144
#define DEF_STRIDE 2
#define DEF_REPEAT 100
#define DEF_FACTOR 2


    if (mysize < 2) {
	printf ("At least 2 processes are required. Aborting.\n");
	MPI_Finalize();	
	exit (1);
    }


    /* default values */
    size = DEF_SIZE;
    bls = DEF_BL_START;
    ble = DEF_BL_END;
    f = DEF_FACTOR;
    stride = DEF_STRIDE;
    repeats = DEF_REPEAT;


    /* args:
       -s size of data to send
       -b blocklen to start with
       -e blocklen to end with
       -f factor to increase the blocklen in each run
       -x stride as multiply of blocklen
       -r nbr of repeats
    */

    while((c = getopt(argc, argv, "s:b:e:f:x:r:?")) != EOF) {
	switch(c) {
	case 's':
	  size = atoi(optarg);
	  break;
	case 'b':
	  bls = atoi(optarg);
	  break;
	case 'e':
	  ble = atoi(optarg);
	  break;
	case 'f':
	  f = atoi(optarg);
	  break;
	case 'x':
	  stride = atoi(optarg);
	  break;
	case 'r':
	  repeats = atoi(optarg);
	  break;
	case '?':
	  if (myrank == 0) {
	  printf ("ncpingpong - test a pingpong with non-contig datatype (vector) \n\
Arguments:\n\
 -s size : overall size of DATA to send (gross amount)\n\
 -b blocklen_start : start test with this blocklen\n\
 -e blocklen_end : end test with this blocklen\n\
 -f factor : factor to increase the blocklen in each run\n\
 -x stride : stride between blocks, in multiply of actual blocklen\n\
 -r repeats : repeat the test <repeats> times\n");
	  }
	  exit (0);
	  break;
	}
    }
    
    if ( myrank == 0) {
	printf ("# Running with this settings: \n\
# size:\t\t%d\n\
# blen_start:\t%d\n\
# blen_end:\t%d\n\
# factor:\t%d\n\
# stride:\t%d\n\
# repeats:\t%d\n\n",size,bls,ble,f,stride,repeats);
	printf ("# Count\tBLlen/(Bytes)\tContig/(MB/s)\tNon-Contig/(MB/s)\tBuffer Size(Bytes)\n");
    }

    for (blocklen = bls; blocklen <= ble; blocklen *= f) {

	count = size / blocklen;

	MPI_Type_vector (count, blocklen, stride*blocklen, MPI_CHAR, &noncontig);
	MPI_Type_commit (&noncontig);
	MPI_Type_contiguous (count*blocklen, MPI_CHAR, &contig);
	MPI_Type_commit (&contig);
	
	MPI_Type_extent (noncontig, (MPI_Aint *)&total_size); 

	buffer = (int *)malloc (total_size);
	recv = (int *)malloc (total_size);
	
	if ( myrank == 0 ) {

	    printf ("%d\t%d\t\t",count,blocklen);
	    
	    wtime = 0.0;
	    for (i = 0; i < repeats; i++) {
		start = MPI_Wtime();
		MPI_Send(buffer, 1, contig, 1, 0, MPI_COMM_WORLD);
		MPI_Recv(recv, 1, contig, 1, 0, MPI_COMM_WORLD, &status);
		wtime += (MPI_Wtime() - start);
	    }
	    bandwidth = (count * blocklen * repeats * 2)/(wtime*1024*1024);
	    latency = wtime / (repeats * 2) * 1e+6;
	    printf ("%f\t", bandwidth);

	    wtime = 0.0;
	    for (i = 0; i < repeats; i++) {
		start = MPI_Wtime();
		MPI_Send(buffer, 1, noncontig, 1, 0, MPI_COMM_WORLD);
		MPI_Recv(recv, 1, noncontig, 1, 0, MPI_COMM_WORLD, &status);
		wtime += (MPI_Wtime() - start);
	    }
	    bandwidth = (count * blocklen * repeats * 2)/(wtime*1024*1024);
	    latency = wtime / (repeats * 2) * 1e+6;
	    printf ("%f\t\t%d\n", bandwidth, total_size); fflush (stdout);
 	} else {
	    if ( myrank == 1 ) {
		for (i = 0; i < repeats; i++) {
		    MPI_Recv(recv, 1, contig, 0, 0, MPI_COMM_WORLD, &status);
		    MPI_Send(recv, 1, contig, 0, 0, MPI_COMM_WORLD);
		}
		for (i = 0; i < repeats; i++) {
		    MPI_Recv(recv, 1, noncontig, 0, 0, MPI_COMM_WORLD, &status);
		    MPI_Send(recv, 1, noncontig, 0, 0, MPI_COMM_WORLD);
		}
	    }
	} 
	MPI_Type_free(&noncontig);
	MPI_Type_free(&contig);
	free (buffer);
	free (recv);
    }
    
    MPI_Finalize();
}


